// Copyright 2018 RnD Center "ELVEES", JSC

#include <mutex>
#include <utility>
#include <vector>

#include <boost/intrusive_ptr.hpp>

#include "elcore-cl.hpp"

std::pair<ecl_int, bool> CheckIfNodeNeedToWaitEvent(ecl_command_queue_node *node,
                                                    _ecl_event *event) {
    bool need_to_wait_before_submit = false;
    ecl_int command_status = ECL_COMPLETE;
    if (node->m_event->m_command_type == ECL_COMMAND_NDRANGE_KERNEL) {
        if (event->m_command_type == ECL_COMMAND_NDRANGE_KERNEL &&
            event->m_queue == node->m_queue) {
            if (event->m_command_status > ECL_SUBMITTED) {
                need_to_wait_before_submit = true;
                command_status = ECL_SUBMITTED;
            }
        } else {
            if (event->m_command_status > ECL_COMPLETE) {
                need_to_wait_before_submit = true;
                command_status = ECL_COMPLETE;
            }
        }
    } else {
        if (event->m_command_status > ECL_COMPLETE) {
            need_to_wait_before_submit = true;
            command_status = ECL_COMPLETE;
        }
    }
    return std::make_pair(command_status, need_to_wait_before_submit);
}

void AdvanceCommandQueueNode(ecl_command_queue_node *node) {
    int ret;
    bool need_to_wait_before_submit = false;
    for (std::size_t i = 0; i < node->m_wait_list.size(); ++i) {
        boost::intrusive_ptr<_ecl_event> event = node->m_wait_list[i];
        std::unique_lock<std::mutex> lock(event->m_mutex);
        need_to_wait_before_submit |= CheckIfNodeNeedToWaitEvent(node, event.get()).second;
    }
    ecl_map_flags flags;
    size_t offset, cb;
    ecl_mem mem;
    if (!need_to_wait_before_submit) {
        switch (node->m_event->m_command_type) {
        case ECL_COMMAND_NDRANGE_KERNEL:
            if (node->use_shared_queue)
                ret = node->m_queue->m_device->enqueue_shared_job(&node->m_kernel->m_job_instance);
            else
                ret = node->m_queue->m_device->enqueue_job(&node->m_kernel->m_job_instance);
            if (ret != 0)
                return;  // ECL_INVALID_CONTEXT;  //@TODO: return correct error
            node->m_event->set_command_status(ECL_SUBMITTED);
            {
                std::unique_lock<std::mutex> lock(node->m_queue->m_commands_list_guard);
                node->m_queue->m_submitted_kernels_list.push_back(node);
                node->m_queue->m_last_command_in_queue = node;
            }
            node->m_queue->m_commands_list_cond.notify_all();
            break;
        case ECL_COMMAND_MAP_BUFFER:
            offset = node->m_sync_object->m_offset;
            cb = node->m_sync_object->m_cb;
            flags = node->m_sync_object->m_map_flags;
            mem = node->m_sync_object->m_memobj;
            if (mem->sync_for_host(offset, cb, flags) != 0)
                return;
            node->m_event->set_command_status(ECL_COMPLETE);
            break;
        case ECL_COMMAND_UNMAP_MEM_OBJECT:
            mem = node->m_sync_object->m_memobj;
            if (mem->sync_for_device() != 0)
                return;
            node->m_event->set_command_status(ECL_COMPLETE);
            break;
        default:
            node->m_event->set_command_status(ECL_COMPLETE);
            return;
        }
    }
}

void ECL_CALLBACK WaitNodeCallback(ecl_event, ecl_int, void *user_data) {
    // @TODO: check errors
    ecl_command_queue_node *node = reinterpret_cast<ecl_command_queue_node *>(user_data);
    AdvanceCommandQueueNode(node);

    {
        std::unique_lock<std::mutex> lock(node->m_queue->m_commands_list_guard);
        if (node->m_queue->m_last_command_in_queue == node)
            node->m_queue->m_last_command_in_queue = nullptr;
    }
    node->release();
}

bool FillWaitListCallback(boost::intrusive_ptr<ecl_command_queue_node> node) {
    bool need_to_wait_before_submit = false;
    for (std::size_t i = 0; i < node->m_wait_list.size(); ++i) {
        boost::intrusive_ptr<_ecl_event> event = node->m_wait_list[i];
        std::unique_lock<std::mutex> lock(event->m_mutex);
        std::pair<ecl_int, bool> check_result = CheckIfNodeNeedToWaitEvent(node.get(), event.get());
        need_to_wait_before_submit |= check_result.second;
        if (check_result.second) {
            node->add_ref();
            event->m_callbacks[check_result.first]
            .push_back(std::pair<void(ECL_CALLBACK *)(ecl_event, ecl_int, void *), void *>(
                       WaitNodeCallback, node.get()));
        }
    }
    return need_to_wait_before_submit;
}

ecl_uint SetInternalArgs(ecl_multi_command_queue command_queue,
                         std::shared_ptr<_ecl_device> device,
                         ecl_kernel kernel, elcore50_job_instance &job_instance) {
    auto context = command_queue->m_context;
    int core_idx = device->m_devnum;
    ecl_uint current_arg_idx = 0;

    int dev_idx = -1;
    for (size_t i = 0; i < context->m_devices.size(); ++i) {
        if (context->m_devices[i] == device)
            dev_idx = i;
    }

    job_instance.args[current_arg_idx].type = ELCORE50_TYPE_NC_GLOBAL_MEMORY;
    job_instance.args[current_arg_idx].global_memory.mapper_fd =
                                    command_queue->m_nc_mem->m_mem_info.mapper_fd;
    current_arg_idx += 1;
    job_instance.args[current_arg_idx].type = ELCORE50_TYPE_GLOBAL_MEMORY;
    job_instance.args[current_arg_idx].global_memory.mapper_fd =
                                    command_queue->m_internal_mem->m_mem_info.mapper_fd;
    current_arg_idx += 1;

    kernel_spawn_internal_data *data =
                (kernel_spawn_internal_data *) command_queue->m_internal_mem->m_host_ptr;
    command_queue->m_internal_mem->sync_for_host(0, sizeof(kernel_spawn_internal_data),
                                         ECL_MAP_READ | ECL_MAP_WRITE);
    data->kernel[core_idx] = kernel->m_kernel_addr;
    data->nc_mem_size = command_queue->m_nc_mem->m_size;
    data->argc[core_idx] = kernel->m_args.size();
    data->context_capacity = command_queue->m_spawn_queues.size() + 1;
    data->index_in_context[core_idx] = dev_idx;
    command_queue->m_internal_mem->sync_for_device();

    return current_arg_idx;
}

static ecl_uint
checkEnqueueParams(ecl_command_queue command_queue, ecl_kernel kernel, ecl_uint work_dim,
                   const size_t *global_work_offset, const size_t *global_work_size,
                   const size_t *local_work_size, ecl_uint num_events_in_wait_list,
                   const ecl_event *event_wait_list, ecl_event *event) {
    if (command_queue == nullptr) return ECL_INVALID_COMMAND_QUEUE;
    if (kernel == nullptr) return ECL_INVALID_KERNEL;
    if (command_queue->m_context != kernel->m_program->m_context) return ECL_INVALID_CONTEXT;

    if (event_wait_list == nullptr && num_events_in_wait_list > 0)
        return ECL_INVALID_EVENT_WAIT_LIST;

    if (event_wait_list != nullptr && num_events_in_wait_list == 0)
        return ECL_INVALID_EVENT_WAIT_LIST;

    if (event_wait_list) {
        for (ecl_uint i = 0; i < num_events_in_wait_list; i++) {
            if (event_wait_list[i] == nullptr) return ECL_INVALID_EVENT_WAIT_LIST;
            if (event_wait_list[i]->m_context != command_queue->m_context)
                return ECL_INVALID_CONTEXT;
        }
    }

    if (work_dim != 1) return ECL_INVALID_WORK_DIMENSION;
    if (global_work_size == nullptr) return ECL_INVALID_GLOBAL_WORK_SIZE;
    if (global_work_size[0] != 1) return ECL_INVALID_GLOBAL_WORK_SIZE;
    if (global_work_offset != nullptr) return ECL_INVALID_GLOBAL_OFFSET;

    {
        std::unique_lock<std::mutex> lock(kernel->m_kernel_guard);
        if (!kernel->m_args.empty()) {
            auto next_it = kernel->m_args.cbegin();
            for (auto it = next_it++; next_it != kernel->m_args.cend(); ++it, ++next_it) {
                if (it->first != (next_it->first - 1)) return ECL_INVALID_KERNEL_ARGS;
            }
        }
    }

    if (local_work_size != nullptr) return ECL_INVALID_WORK_ITEM_SIZE;
    // @TODO: check if we can map memories to virtual space
    // else return ECL_MEM_OBJECT_ALLOCATION_FAILURE

    return ECL_SUCCESS;
}

static ecl_uint
eclEnqueueKernelInternal(ecl_command_queue command_queue, ecl_kernel kernel, ecl_uint work_dim,
                         const size_t *global_work_offset, const size_t *global_work_size,
                         const size_t *local_work_size, ecl_uint num_events_in_wait_list,
                         const ecl_event *event_wait_list, ecl_event *event,
                         ecl_multi_command_queue spawn_multi_queue, ecl_bool use_shared_queue) {
    ecl_uint ret;

    ret = checkEnqueueParams(command_queue, kernel, work_dim, global_work_offset, global_work_size,
                             local_work_size, num_events_in_wait_list, event_wait_list, event);
    if (ret != ECL_SUCCESS) return ret;

    const boost::intrusive_ptr<_ecl_program> program = kernel->m_program;

    int dev_num = -1;
    for (size_t dev = 0; dev < program->m_devices.size(); dev++) {
        if (program->m_devices[dev] == command_queue->m_device_id) {
            dev_num = dev;
            break;
        }
    }
    if (dev_num == -1) {
        return ECL_INVALID_CONTEXT;
    }

    boost::intrusive_ptr<ecl_command_queue_node> node = new ecl_command_queue_node();
    node->m_kernel = new _ecl_kernel_instance();
    node->m_kernel->m_job_instance = program->m_job_instance_templates[dev_num];
    elcore50_job_instance &job_instance = node->m_kernel->m_job_instance;

    {
        std::unique_lock<std::mutex> lock(kernel->m_kernel_guard);

        ecl_uint internal_argc = 0;
        if (spawn_multi_queue) {
            /* Set internal args for elcore-runtime */
            internal_argc = SetInternalArgs(spawn_multi_queue, command_queue->m_device,
                                            kernel, job_instance);
        }

        job_instance.argc = kernel->m_args.size() + internal_argc;
        for (auto it = kernel->m_args.cbegin(); it != kernel->m_args.cend(); ++it) {
            const ecl_uint index = internal_argc + it->first;
            const ecl_uint type = it->second.first;
            switch (type) {
                case ECL_KERNEL_ARG_TYPE_BASIC: {
                    job_instance.args[index].type = ELCORE50_TYPE_BASIC;
                    node->m_basic_args_copy.push_back(it->second.second);
                    const std::vector<uint8_t> &data = node->m_basic_args_copy.back();
                    job_instance.args[index].basic.p = reinterpret_cast<uintptr_t>(data.data());
                    job_instance.args[index].basic.size = data.size();
                    break;
                }
                case ECL_KERNEL_ARG_TYPE_NC_GLOBAL:
                case ECL_KERNEL_ARG_TYPE_GLOBAL: {
                    if (type == ECL_KERNEL_ARG_TYPE_GLOBAL)
                        job_instance.args[index].type = ELCORE50_TYPE_GLOBAL_MEMORY;
                    else
                        job_instance.args[index].type = ELCORE50_TYPE_NC_GLOBAL_MEMORY;
                    node->m_mem_args_ref.emplace_back(
                        *reinterpret_cast<const ecl_mem *>(it->second.second.data()));
                    ecl_mem mem = node->m_mem_args_ref.back().get();
                    job_instance.args[index].global_memory.mapper_fd = mem->m_mem_info.mapper_fd;
                    break;
                }
                case ECL_KERNEL_ARG_TYPE_LOCAL:
                    job_instance.args[index].type = ELCORE50_TYPE_LOCAL_MEMORY;
                    job_instance.args[index].local_memory.size =
                        *reinterpret_cast<const size_t *>(it->second.second.data());
                    break;
                case ECL_KERNEL_ARG_TYPE_DMA: {
                    job_instance.args[index].type = ELCORE50_TYPE_DMA_MEMORY;
                    node->m_mem_args_ref.emplace_back(
                        *reinterpret_cast<const ecl_mem *>(it->second.second.data()));
                    ecl_mem mem = node->m_mem_args_ref.back().get();
                    job_instance.args[index].dma_memory.mapper_fd = mem->m_mem_info.mapper_fd;
                    break;
                }
            }
        }
    }
    job_instance.entry_point_virtual_address =
        (spawn_multi_queue == nullptr) ? kernel->m_kernel_addr :
            kernel->m_program->m_elf->GetKernelVirtualAddress("_elcore_kernel_wrapper");
    strncpy(job_instance.name, kernel->m_kernel_name.c_str(), sizeof(job_instance.name) - 1);
    job_instance.debug_enable = program->m_debug_enable;

    node->m_event =
        new _ecl_event(command_queue->m_context.get(), command_queue, ECL_COMMAND_NDRANGE_KERNEL);
    node->m_queue = command_queue;
    node->use_shared_queue = use_shared_queue;
    if (num_events_in_wait_list) {
        node->m_wait_list.assign(event_wait_list, event_wait_list + num_events_in_wait_list);
    }
    {
        std::unique_lock<std::mutex> lock(command_queue->m_commands_list_guard);
        if (command_queue->m_last_command_in_queue != nullptr) {
            node->m_wait_list.push_back(command_queue->m_last_command_in_queue->m_event);
        }
    }
    if (!FillWaitListCallback(node)) {
        if (use_shared_queue)
            ret = command_queue->m_device->enqueue_shared_job(&node->m_kernel->m_job_instance);
        else
            ret = command_queue->m_device->enqueue_job(&node->m_kernel->m_job_instance);
        if (ret != 0) {
            return ECL_INVALID_CONTEXT;  // @TODO: return correct error
        }
        node->m_event->set_command_status(ECL_SUBMITTED);
        {
            std::unique_lock<std::mutex> lock(command_queue->m_commands_list_guard);
            command_queue->m_submitted_kernels_list.push_back(node);
            command_queue->m_last_command_in_queue = node;
        }
        command_queue->m_commands_list_cond.notify_all();
    }
    if (event) {
        node->m_event->add_ref();
        *event = node->m_event.get();
    }
    return ECL_SUCCESS;
}

extern "C" ECL_API_ENTRY ecl_int ECL_API_CALL
eclEnqueueNDRangeKernel(ecl_command_queue command_queue, ecl_kernel kernel, ecl_uint work_dim,
                        const size_t *global_work_offset, const size_t *global_work_size,
                        const size_t *local_work_size, ecl_uint num_events_in_wait_list,
                        const ecl_event *event_wait_list,
                        ecl_event *event) ECL_API_SUFFIX__VERSION_1_0 {
    return eclEnqueueKernelInternal(command_queue, kernel, work_dim, global_work_offset,
                                    global_work_size, local_work_size, num_events_in_wait_list,
                                    event_wait_list, event, nullptr, ECL_FALSE);
}

extern "C" ECL_API_ENTRY ecl_int ECL_API_CALL
eclEnqueueSharedKernel(ecl_command_queue command_queue, ecl_kernel kernel, ecl_uint work_dim,
                       const size_t *global_work_offset, const size_t *global_work_size,
                       const size_t *local_work_size, ecl_uint num_events_in_wait_list,
                       const ecl_event *event_wait_list,
                       ecl_event *event) ECL_API_SUFFIX__VERSION_1_0 {
    return eclEnqueueKernelInternal(command_queue, kernel, work_dim, global_work_offset,
                                    global_work_size, local_work_size, num_events_in_wait_list,
                                    event_wait_list, event, nullptr, ECL_TRUE);
}

extern "C" ECL_API_ENTRY ecl_int ECL_API_CALL
eclEnqueueKernelWithSpawn(ecl_multi_command_queue command_queue, ecl_kernel kernel,
                          ecl_uint num_events_in_wait_list, const ecl_event *event_wait_list,
                          ecl_event *event) ECL_API_SUFFIX__VERSION_1_0 {
    ecl_int ret;
    ecl_program program = kernel->m_program.get();
    size_t localmem_size = 0;

    for (auto it = kernel->m_args.cbegin(); it != kernel->m_args.cend(); ++it) {
        const ecl_uint type = it->second.first;
        // Make sure there are no ECL_KERNEL_ARG_TYPE_BASIC arguments with size > 8 bytes
        if (type == ECL_KERNEL_ARG_TYPE_BASIC) {
            if (it->second.second.size() > 8)
                return ECL_INVALID_KERNEL_ARGS;
        }
        // Calculate passed localmem size for kernel
        if (type == ECL_KERNEL_ARG_TYPE_LOCAL)
            localmem_size += *reinterpret_cast<const size_t *>(it->second.second.data());
    }

    ecl_kernel spawn_kernel = eclCreateKernel(program, "_spawner_loop", &ret);
    if (ret != ECL_SUCCESS) return ret;

    if (localmem_size) {
        // Pass localmem to spawn_kernel if the main kernel uses localmem
        ret = eclSetKernelArgELcoreLocalMem(spawn_kernel, 0, localmem_size);
        if (ret != ECL_SUCCESS) {
            eclReleaseKernel(spawn_kernel);
            return ret;
        }
    }

    /* TODO: Fix error handling */
    /* Если:
          1) spawn_kernel не удалось запутить на одном из девайсов (не самом первом);
          2) не удалось запустить основной кернел при запущенных spawn_kernels;
          3) один из spawn_kernel по какой-то причине или завис или не завершился к
             моменту завершения основного кернела,
       то мы не сможем корректно завершить job_instance'ы и выдать правильную ошибку.
       Для исправления первых двух ошибок в текущий реализации используются исключения
       для принудительного завершения всего процесса.
    */

    const size_t global_work_size[1] = {1};
    for (size_t i = 0; i < command_queue->m_spawn_queues.size(); ++i) {
        ret = eclEnqueueKernelInternal(command_queue->m_spawn_queues[i], spawn_kernel, 1, nullptr,
                                       global_work_size, nullptr, 0, nullptr, nullptr,
                                       command_queue, ECL_FALSE);
        if (ret != ECL_SUCCESS) throw std::runtime_error("Failed to start spawn kerenel");
    }

    ret = eclEnqueueKernelInternal(command_queue, kernel, 1, nullptr, global_work_size, nullptr,
                                   num_events_in_wait_list, event_wait_list, event, command_queue,
                                   ECL_FALSE);
    if (ret != ECL_SUCCESS) throw std::runtime_error("Failed to start kernel");

    eclReleaseKernel(spawn_kernel);

    return ECL_SUCCESS;
}

extern "C" ECL_API_ENTRY void * ECL_API_CALL
eclEnqueueMapBuffer(ecl_command_queue command_queue, ecl_mem memobj, ecl_bool blocking_map,
                    ecl_map_flags map_flags, size_t offset, size_t cb,
                    ecl_uint num_events_in_wait_list, const ecl_event *event_wait_list,
                    ecl_event *event, ecl_int *errcode_ret) ECL_API_SUFFIX__VERSION_1_0 {
    if (command_queue == nullptr) {
        *errcode_ret = ECL_INVALID_COMMAND_QUEUE;
        return NULL;
    }

    if (memobj == nullptr) {
        *errcode_ret = ECL_INVALID_MEM_OBJECT;
        return NULL;
    }

    if (memobj->m_host_ptr == nullptr) {
        *errcode_ret = ECL_INVALID_OPERATION;
        return NULL;
    }

    if ((cb == 0) || (offset + cb > memobj->m_size)) {
        *errcode_ret = ECL_INVALID_ARG_SIZE;
        return NULL;
    }

    if (map_flags & ~(ECL_MAP_READ | ECL_MAP_WRITE)) {
        *errcode_ret =  ECL_INVALID_VALUE;
        return NULL;
    }

    if (event_wait_list == nullptr && num_events_in_wait_list > 0) {
        *errcode_ret = ECL_INVALID_EVENT_WAIT_LIST;
        return NULL;
    }

    if (event_wait_list != nullptr && num_events_in_wait_list == 0) {
        *errcode_ret = ECL_INVALID_EVENT_WAIT_LIST;
        return NULL;
    }

    if (event_wait_list) {
        for (ecl_uint i = 0; i < num_events_in_wait_list; i++) {
            if (event_wait_list[i] == nullptr) {
                *errcode_ret = ECL_INVALID_EVENT_WAIT_LIST;
                return NULL;
            }
            if (event_wait_list[i]->m_context != command_queue->m_context) {
                *errcode_ret = ECL_INVALID_CONTEXT;
                return NULL;
            }
        }
    }

    boost::intrusive_ptr<ecl_command_queue_node> node = new ecl_command_queue_node();
    node->m_event =
        new _ecl_event(command_queue->m_context.get(), command_queue, ECL_COMMAND_MAP_BUFFER);
    node->m_queue = command_queue;
    if (num_events_in_wait_list) {
        node->m_wait_list.assign(event_wait_list, event_wait_list + num_events_in_wait_list);
    }
    {
        std::unique_lock<std::mutex> lock(command_queue->m_commands_list_guard);
        if (command_queue->m_last_command_in_queue != nullptr) {
            node->m_wait_list.push_back(command_queue->m_last_command_in_queue->m_event);
        }
    }
    boost::intrusive_ptr<_ecl_sync_object_instance> sync = new _ecl_sync_object_instance();
    sync->m_memobj = memobj;
    sync->m_map_flags = map_flags;
    sync->m_cb = cb;
    sync->m_offset = offset;

    node->m_sync_object = sync;

    if (!FillWaitListCallback(node)) {
        if (memobj->sync_for_host(offset, cb, map_flags) != 0) {
            *errcode_ret = ECL_DEVICE_NOT_AVAILABLE;
            return NULL;
        }
        node->m_event->set_command_status(ECL_COMPLETE);
    }
    if (event) {
        node->m_event->add_ref();
        *event = node->m_event.get();
    }

    if (blocking_map) {
        ecl_event block_event = node->m_event.get();
        eclWaitForEvents(1, &block_event);
    }

    if (errcode_ret) *errcode_ret = ECL_SUCCESS;
    return reinterpret_cast<void *>(reinterpret_cast<uintptr_t>(memobj->m_host_ptr) + offset);
}

extern "C" ECL_API_ENTRY ecl_int ECL_API_CALL
eclEnqueueUnmapMemObject(ecl_command_queue command_queue, ecl_mem memobj, void *mapped_ptr,
                         ecl_uint num_events_in_wait_list, const ecl_event *event_wait_list,
                         ecl_event *event) ECL_API_SUFFIX__VERSION_1_0 {
    if (command_queue == nullptr) return ECL_INVALID_COMMAND_QUEUE;
    if (memobj == nullptr) return ECL_INVALID_MEM_OBJECT;

    if (event_wait_list == nullptr && num_events_in_wait_list > 0)
        return ECL_INVALID_EVENT_WAIT_LIST;

    if (event_wait_list != nullptr && num_events_in_wait_list == 0)
        return ECL_INVALID_EVENT_WAIT_LIST;

    if (event_wait_list) {
        for (ecl_uint i = 0; i < num_events_in_wait_list; i++) {
            if (event_wait_list[i] == nullptr) return ECL_INVALID_EVENT_WAIT_LIST;
            if (event_wait_list[i]->m_context != command_queue->m_context)
                return ECL_INVALID_CONTEXT;
        }
    }

    boost::intrusive_ptr<ecl_command_queue_node> node = new ecl_command_queue_node();
    node->m_event =
        new _ecl_event(command_queue->m_context.get(), command_queue,
                       ECL_COMMAND_UNMAP_MEM_OBJECT);
    node->m_queue = command_queue;
    if (num_events_in_wait_list) {
        node->m_wait_list.assign(event_wait_list, event_wait_list + num_events_in_wait_list);
    }
    {
        std::unique_lock<std::mutex> lock(command_queue->m_commands_list_guard);
        if (command_queue->m_last_command_in_queue != nullptr) {
            node->m_wait_list.push_back(command_queue->m_last_command_in_queue->m_event);
        }
    }
    boost::intrusive_ptr<_ecl_sync_object_instance> sync = new _ecl_sync_object_instance();
    sync->m_memobj = memobj;

    node->m_sync_object = sync;

    if (!FillWaitListCallback(node)) {
        if (memobj->sync_for_device() != 0) {
            return ECL_DEVICE_NOT_AVAILABLE;
        }
        node->m_event->set_command_status(ECL_COMPLETE);
    }
    if (event) {
        node->m_event->add_ref();
        *event = node->m_event.get();
    }

    return ECL_SUCCESS;
}
