// Copyright 2021 RnD Center "ELVEES", JSC

#include <fcntl.h>
#include <fstream>
#include <gtest/gtest.h>
#include <poll.h>
#include <pthread.h>
#include <sys/ioctl.h>

#include "libelcore.hpp"
#ifdef DRIVER_NOT_INSTALLED
#include <elcore50.h>
#else
#include <linux/elcore50.h>
#endif

class ELcoreCounter {
 public:
    ELcoreCounter() {
        int m_fd = open("/dev/elcore0", O_RDONLY, 0);
        if (m_fd == -1) throw std::runtime_error("can't open device");

        struct elcore50_device_info device_info;
        ioctl(m_fd, ELCORE50_IOC_GET_CORE_IDX, &device_info);
        close(m_fd);

        m_elcore_num = device_info.nclusters * device_info.cluster_cap;
    }
    int m_elcore_num;
};

struct pthread_data {
    int job_instance_fd;
    int fd;
};

void *poll_job(void *args) {
    struct pthread_data *pdata = (struct pthread_data *) args;

    int status = ElcoreJobInstancePoll(pdata->job_instance_fd, pdata->fd);
    pthread_exit((void *)(status != ELCORE50_JOB_STATUS_SUCCESS));
}

ELcoreCounter g_elcore_counter;

class JobTests : public ::testing::TestWithParam<int> {
 public:
    JobTests()  {}

 protected:
    void SetUp() override {
        m_core_num = GetParam();

        std::vector<char> elf_buffer;
        size_t elf_size;
        {
            std::ifstream file(TEST_ELF_FILE, std::ios::binary | std::ios::ate);
            elf_size = file.tellg();
            file.seekg(0, std::ios::beg);
            elf_buffer.resize(elf_size);
            ASSERT_TRUE(file.read(elf_buffer.data(), elf_size));
        }

        std::shared_ptr<ELFParser> parser =
            CreateELFParser((const unsigned char *) elf_buffer.data(), elf_size);

        m_elf = ElcoreELF::CreateELF(parser);
        ASSERT_TRUE(m_elf != nullptr);

        char devname[255];

        sprintf(devname, "/dev/elcore%d", m_core_num);
        m_fd = open(devname, O_RDWR, 0);
        ASSERT_TRUE(m_fd > 0);

        m_job = ElcoreJob::CreateElcoreJob(parser, m_elf, m_fd);
        ASSERT_TRUE(m_job != nullptr);

        m_job_instance.job_fd = m_job->GetFD();
        m_job_instance.entry_point_virtual_address = m_elf->GetEntryPoint();
        m_job_instance.launcher_virtual_address =
            m_elf->GetKernelVirtualAddress("_elcore_job_launcher");
        m_job_instance.argc = 0;
        m_job_instance.debug_enable = 0;
        strcpy(m_job_instance.name, "__start");

        ASSERT_EQ(ioctl(m_fd, ELCORE50_IOC_ENQUEUE_JOB, &m_job_instance), 0);
        int status = ElcoreJobInstancePoll(m_job_instance.job_instance_fd, m_fd);
        ASSERT_EQ(status, ELCORE50_JOB_STATUS_SUCCESS);
        close(m_job_instance.job_instance_fd);
    }

    void TearDown() override {
        if (m_job_instance.job_instance_fd)
            close(m_job_instance.job_instance_fd);
        if (m_fd)
            close(m_fd);
    }

    void CreateJobInstance(const char *kernel_name, int debug = 0) {
        m_job_instance.entry_point_virtual_address = m_elf->GetKernelVirtualAddress(kernel_name);
        m_job_instance.launcher_virtual_address =
            m_elf->GetKernelVirtualAddress("_elcore_job_launcher");
        m_job_instance.argc = 0;
        m_job_instance.debug_enable = debug;
        strncpy(m_job_instance.name, kernel_name, sizeof(m_job_instance.name) - 1);

        // TODO: Set args

        ASSERT_EQ(ioctl(m_fd, ELCORE50_IOC_ENQUEUE_JOB, &m_job_instance), 0);
        ASSERT_TRUE(m_job_instance.job_instance_fd != 0);
        pdata.job_instance_fd = m_job_instance_fd = m_job_instance.job_instance_fd;
        pdata.fd = m_fd;
        ASSERT_EQ(pthread_create(&m_poll_thread, NULL, poll_job, (void *) &pdata), 0);
    }

    void WaitJob() {
        int status_addr, status;
        status = pthread_join(m_poll_thread, (void**)&status_addr);
        ASSERT_EQ(status, 0);
        ASSERT_EQ(status_addr, 0);
    }

    long WaitJobInstanceCreation(const char *name, int timeout_sec) {
        struct elcore50_job_instance_list list;
        __u32 job_count;
        /* Wait for kernel,  */
        int timeout = timeout_sec * 1000;  // ms
        int time_step = 10;  // 10 ms
        long ret;
        while (timeout) {
            if (ioctl(m_fd, ELCORE50_IOC_GET_JOB_COUNT, &job_count) < 0) {
                fprintf(stderr, "ELCORE50_IOC_GET_JOB_COUNT failed with %d\n", errno);
                return 0;
            }
            if (!job_count) {
                usleep(time_step);
                timeout -= time_step;
                continue;
            }
            list.info = (struct elcore50_job_instance_info *)
                    malloc(job_count * sizeof(struct elcore50_job_instance_info));
            if (!list.info) {
                fprintf(stderr, "Failed to allocate memory\n");
                return 0;
            }

            list.job_instance_count = job_count;
            list.job_instance_ret = 0;

            if (ioctl(m_fd, ELCORE50_IOC_GET_JOB_LIST, &list) < 0) {
                fprintf(stderr, "ELCORE50_IOC_GET_JOB_LIST failed with %d\n", errno);
                return 0;
            }
            for (uint32_t i = 0; i < list.job_instance_ret; ++i) {
                if (!strcmp(name, list.info[i].name)) {
                    ret = list.info[i].id;
                    free(list.info);
                    return ret;
                }
            }
            free(list.info);
            usleep(time_step);
            timeout -= time_step;
        }
        return 0;
    }

    int AttachJob(long job_instance_id) {
        m_job_inst_dbg.job_instance_id = job_instance_id;
        if (ioctl(m_fd, ELCORE50_IOC_DBG_JOB_ATTACH, &m_job_inst_dbg) < 0) {
            fprintf(stderr, "ELCORE50_IOC_DBG_JOB_ATTACH failed with %d\n", errno);
            m_job_inst_dbg.job_instance_dbg_fd = 0;
        }
        return m_job_inst_dbg.job_instance_dbg_fd;
    }

    void DetachJob(int debug_fd) {
        ASSERT_EQ(m_job_inst_dbg.job_instance_dbg_fd, debug_fd);
        close(debug_fd);
    }

    void DebugPoll(int fd) {
        elcore50_job_instance_status job_status;
        struct pollfd fds;

        job_status.job_instance_fd = m_job_instance_fd;
        fds.fd = fd;
        fds.events = POLLIN;

        while (1) {
            if (poll(&fds, 1, -1) <= 0) {
                throw std::runtime_error("poll errored out or timed out");
            }
            if (fds.revents & (POLLERR | POLLNVAL)) {
                throw std::runtime_error("poll returned error");
            }
            ioctl(m_fd, ELCORE50_IOC_GET_JOB_STATUS, &job_status);
            if (job_status.state == ELCORE50_JOB_STATUS_INTERRUPTED)
                break;
        }
    }

    void CreateSyncFileAndWait(const char *filename) {
        /* Create file */
        int fd = open(filename, O_RDWR | O_APPEND | O_CREAT, 0600);
        WaitJob();
        close(fd);
        // remove file
        unlink(filename);
    }

    int m_fd;
    int m_job_instance_fd;
    elcore50_job_instance m_job_instance;
    std::shared_ptr<ElcoreELF> m_elf;

 private:
    std::shared_ptr<ElcoreJob> m_job;
    int m_core_num;
    pthread_t m_poll_thread;
    struct pthread_data pdata;
    struct elcore50_job_instance_dbg m_job_inst_dbg;
};

TEST_P(JobTests, WaitJob) {
    const char kernel_name[255] = "_kernel_wait_for_file";

    CreateJobInstance(kernel_name);

    ASSERT_NE(WaitJobInstanceCreation(kernel_name, 120), 0);

    CreateSyncFileAndWait("/tmp/filename");
}

TEST_P(JobTests, CheckDbgAttachment) {
    const char kernel_name[255] = "_kernel_wait_for_file";

    CreateJobInstance(kernel_name);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);
    DetachJob(debug_fd);

    CreateSyncFileAndWait("/tmp/filename");
}

TEST_P(JobTests, CheckDbgDoubleAttachment) {
    const char kernel_name[255] = "_kernel_wait_for_file";

    CreateJobInstance(kernel_name);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    struct elcore50_job_instance_dbg job_inst_dbg1;
    job_inst_dbg1.job_instance_id = job_instance_id;
    ASSERT_EQ(ioctl(m_fd, ELCORE50_IOC_DBG_JOB_ATTACH, &job_inst_dbg1), 0);
    ASSERT_NE(job_inst_dbg1.job_instance_dbg_fd, 0);

    struct elcore50_job_instance_dbg job_inst_dbg2;
    job_inst_dbg2.job_instance_id = job_instance_id;
    ASSERT_NE(ioctl(m_fd, ELCORE50_IOC_DBG_JOB_ATTACH, &job_inst_dbg2), 0);

    close(job_inst_dbg1.job_instance_dbg_fd);

    CreateSyncFileAndWait("/tmp/filename");
}

TEST_P(JobTests, CheckDbgAttachmentFromAnotherProcess) {
    int result;
    const char kernel_name[255] = "_kernel_wait_for_file";

    CreateJobInstance(kernel_name);

    int pid = fork();
    if (pid == 0) {
        long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);

        struct elcore50_job_instance_dbg job_inst_dbg;
        job_inst_dbg.job_instance_id = job_instance_id;
        result = ioctl(m_fd, ELCORE50_IOC_DBG_JOB_ATTACH, &job_inst_dbg);
        if (result < 0) exit(1);
        if (job_inst_dbg.job_instance_dbg_fd == 0) exit(1);

        close(job_inst_dbg.job_instance_dbg_fd);
        exit(0);
    }
    waitpid(pid, &result, 0);
    ASSERT_EQ(result, 0);

    CreateSyncFileAndWait("/tmp/filename");
}

TEST_P(JobTests, ReadAndWriteElcoreGlobalMemory) {
    const char kernel_name[255] = "_kernel_wait_for_data";
    // Read DSP virtual address 0x30000000, expected value: 0xAEBFC00F
    CreateJobInstance(kernel_name);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    elcore50_job_instance_status job_status;
    job_status.job_instance_fd = m_job_instance_fd;
    do {
        ioctl(m_fd, ELCORE50_IOC_GET_JOB_STATUS, &job_status);
    } while (job_status.state != ELCORE50_JOB_STATUS_RUN);

    // DSP should have time to write it to memory
    usleep(1000);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);

    struct elcore50_dbg_mem dbg_mem;
    uint32_t data;
    dbg_mem.vaddr = 0x30000000;
    dbg_mem.size = sizeof(uint32_t);
    dbg_mem.data = &data;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_READ, &dbg_mem), 0);
    ASSERT_EQ(data, 0xAEBFC00F);

    // Write any value to DSP virtual address 0x30000000 and wait job
    data = 0xFFFFFFFF;
    dbg_mem.vaddr = 0x30000000;
    dbg_mem.size = sizeof(uint32_t);
    dbg_mem.data = &data;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_WRITE, &dbg_mem), 0);

    // Detach
    DetachJob(debug_fd);

    WaitJob();
}

TEST_P(JobTests, ReadAndWriteElcoreLocalMemory) {
    const char kernel_name[255] = "_kernel_wait_for_xyram_data";
    CreateJobInstance(kernel_name);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    elcore50_job_instance_status job_status;
    job_status.job_instance_fd = m_job_instance_fd;
    do {
        ioctl(m_fd, ELCORE50_IOC_GET_JOB_STATUS, &job_status);
    } while (job_status.state != ELCORE50_JOB_STATUS_RUN);

    // DSP should have time to write it to memory
    usleep(1000);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);

    struct elcore50_dbg_mem dbg_mem;
    uint32_t data;
#ifdef BUILD_FOR_MCOM03
    dbg_mem.vaddr = 0x2E00000;
#else
    dbg_mem.vaddr = 0x10000000;
#endif
    dbg_mem.size = sizeof(uint32_t);
    dbg_mem.data = &data;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_READ, &dbg_mem), 0);
    ASSERT_EQ(data, 0xAAAABBBB);

#ifdef BUILD_FOR_MCOM03
    dbg_mem.vaddr = 0x2E00000;
#else
    dbg_mem.vaddr = 0x10000000;
#endif
    data = 0xFFEEDDCC;
    dbg_mem.size = sizeof(uint32_t);
    dbg_mem.data = &data;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_WRITE, &dbg_mem), 0);

    // Detach
    DetachJob(debug_fd);

    WaitJob();
}

TEST_P(JobTests, StopAndContinueKernel) {
    const char kernel_name[255] = "_kernel_empty";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);

    // Check if DSP is interrupted by attach
    DebugPoll(debug_fd);

    // Chek stop reason
    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_DBG_INTERRUPT);

    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Detach
    DetachJob(debug_fd);

    WaitJob();
}

TEST_P(JobTests, SoftwareBreakPoint) {
    const char kernel_name[255] = "_kernel_with_simple_recursion";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);

    // Check if DSP is interrupted by attach
    DebugPoll(debug_fd);

    // Chek stop reason
    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_DBG_INTERRUPT);

    // Set dbBreak
    // 1. Save the old instruction
    struct elcore50_dbg_mem dbg_mem;
    uint32_t old_instruction;
    dbg_mem.vaddr = m_elf->GetKernelVirtualAddress("_simple_recursion");
    dbg_mem.size = sizeof(uint32_t);  // size of dbBreak instruction
    dbg_mem.data = &old_instruction;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_READ, &dbg_mem), 0);
    // 2. Write dbBreak
    uint32_t new_instruction = 0x100347;  // dbBreak 0x10
    dbg_mem.data = &new_instruction;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_WRITE, &dbg_mem), 0);

    // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Wait when software breakpoint is triggered
    DebugPoll(debug_fd);

    // Chek stop reason
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_SW_BREAKPOINT);

    // Restore old_instruction
    dbg_mem.data = &old_instruction;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_WRITE, &dbg_mem), 0);

    // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Detach
    DetachJob(debug_fd);

    WaitJob();
}

TEST_P(JobTests, HardwareBreakPoint) {
    const char kernel_name[255] = "_kernel_with_simple_recursion";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);
    DebugPoll(debug_fd);

    // Check stop reason
    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_DBG_INTERRUPT);

    // Set hardware breakpoint
    uint32_t breakpoint_vaddr = m_elf->GetKernelVirtualAddress("_simple_recursion");
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_HW_BREAKPOINT_SET, &breakpoint_vaddr), 0);

    for (int i = 11; i >= 0; --i) {
        // Run DSP
        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

        // Wait when HW breakpoint is triggered.
        DebugPoll(debug_fd);
        if (i == 0) break;

        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
        ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_HW_BREAKPOINT);
    }

    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_APP_EXCEPTION);

    // Detach
    DetachJob(debug_fd);
    WaitJob();
}

TEST_P(JobTests, HardwareAndSoftwareBreakPoints) {
    const char kernel_name[255] = "_kernel_with_labels";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);
    DebugPoll(debug_fd);

    // Check stop reason
    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_DBG_INTERRUPT);

    // Set hardware breakpoint
    uint32_t bp_address[2];
    bp_address[0] = m_elf->GetKernelVirtualAddress("_kernel_with_labels");
    bp_address[1] = bp_address[0] + 4;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_HW_BREAKPOINT_SET, &bp_address[0]), 0);

    // Set software breakpoint
    // 1. Save the old instruction
    struct elcore50_dbg_mem dbg_mem;
    uint32_t old_instruction;
    dbg_mem.vaddr = bp_address[1];
    dbg_mem.size = sizeof(uint32_t);  // size of dbBreak instruction
    dbg_mem.data = &old_instruction;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_READ, &dbg_mem), 0);
    // 2. Write dbBreak
    uint32_t new_instruction = 0x100347;  // dbBreak 0x10
    dbg_mem.data = &new_instruction;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_WRITE, &dbg_mem), 0);

    for (int i = 0; i < 2; ++i) {
        // Run DSP
        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

        // Wait when HW breakpoint is triggered.
        DebugPoll(debug_fd);

        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
        if (i == 0) {
            ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_HW_BREAKPOINT);
        }
        if (i == 1) {
            ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_SW_BREAKPOINT);
        }

        // Read old value of DSP_PC
        uint32_t pc_value;
        struct elcore50_dbg_mem pc_mem;
        pc_mem.vaddr = 0x300;
        pc_mem.size = sizeof(uint32_t);
        pc_mem.data = &pc_value;
        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_REGISTER_READ, &pc_mem), 0);

        ASSERT_EQ(pc_value, bp_address[i]);
    }

    // Restore old_instruction
    dbg_mem.data = &old_instruction;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_MEMORY_WRITE, &dbg_mem), 0);

     // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    DebugPoll(debug_fd);

    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_APP_EXCEPTION);

    // Detach
    DetachJob(debug_fd);
    WaitJob();
}

TEST_P(JobTests, DoubleHardwareBreakPoints) {
    const char kernel_name[255] = "_kernel_with_labels";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);
    DebugPoll(debug_fd);

    // Check stop reason
    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_DBG_INTERRUPT);

    // Set hardware breakpoint
    uint32_t bp_address[2];
    bp_address[0] = m_elf->GetKernelVirtualAddress("_kernel_with_labels");
    bp_address[1] = bp_address[0] + 4;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_HW_BREAKPOINT_SET, &bp_address[0]), 0);
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_HW_BREAKPOINT_SET, &bp_address[1]), 0);

    for (int i = 0; i < 2; ++i) {
        // Run DSP
        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

        // Wait when HW breakpoint is triggered.
        DebugPoll(debug_fd);

        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
        ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_HW_BREAKPOINT);

        // Read old value of DSP_PC
        uint32_t pc_value;
        struct elcore50_dbg_mem dbg_mem;
        dbg_mem.vaddr = 0x300;
        dbg_mem.size = sizeof(uint32_t);
        dbg_mem.data = &pc_value;
        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_REGISTER_READ, &dbg_mem), 0);

        ASSERT_EQ(pc_value, bp_address[i]);
    }

     // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    DebugPoll(debug_fd);

    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_APP_EXCEPTION);

    // Detach
    DetachJob(debug_fd);
    WaitJob();
}

TEST_P(JobTests, CheckDbgStep) {
    const char kernel_name[255] = "_kernel_with_simple_recursion";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);
    DebugPoll(debug_fd);

    // Check stop reason
    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_DBG_INTERRUPT);

    // Read old value of DSP_INSTR_CNTR
    struct elcore50_dbg_mem dbg_mem;
    uint32_t tic_instr_old;
    dbg_mem.vaddr = 0x298;
    dbg_mem.size = sizeof(uint32_t);
    dbg_mem.data = &tic_instr_old;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_REGISTER_READ, &dbg_mem), 0);

    // Execute 10 steps
    uint32_t steps = 10;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_STEP, &steps), 0);

    // Wait when breakpoint is triggered.
    DebugPoll(debug_fd);

    // Chek STOP reason
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_STEP);

    // Read old value of DSP_INSTR_CNTR
    uint32_t tic_instr_new;
    dbg_mem.vaddr = 0x298;
    dbg_mem.size = sizeof(uint32_t);
    dbg_mem.data = &tic_instr_new;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_REGISTER_READ, &dbg_mem), 0);

    // Compare DSP_INSTR_CNTR
    ASSERT_EQ(tic_instr_new, tic_instr_old + steps);

    // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Detach
    DetachJob(debug_fd);
    WaitJob();
}

TEST_P(JobTests, CheckDbgStepAfterBreakpoint) {
    const char kernel_name[255] = "_kernel_with_labels";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);
    DebugPoll(debug_fd);

    // Check stop reason
    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_DBG_INTERRUPT);

    uint32_t bp_address;
    bp_address = m_elf->GetKernelVirtualAddress("_kernel_with_labels");
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_HW_BREAKPOINT_SET, &bp_address), 0);

    // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Wait when HW breakpoint is triggered.
    DebugPoll(debug_fd);

    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_HW_BREAKPOINT);

    // Execute 1 step
    uint32_t steps = 1;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_STEP, &steps), 0);

    // Wait when breakpoint is triggered.
    DebugPoll(debug_fd);

    // Chek STOP reason
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_STEP);

    uint32_t pc_value;
    struct elcore50_dbg_mem pc_mem;
    pc_mem.vaddr = 0x300;
    pc_mem.size = sizeof(uint32_t);
    pc_mem.data = &pc_value;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_REGISTER_READ, &pc_mem), 0);
    ASSERT_EQ(pc_value, bp_address + 4);

    // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Detach
    DetachJob(debug_fd);
    WaitJob();
}

TEST_P(JobTests, CheckDbgStepOverStop) {
    const char kernel_name[255] = "_kernel_with_stop";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);
    DebugPoll(debug_fd);

    // Check stop reason
    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_DBG_INTERRUPT);

    uint32_t bp_address;
    bp_address = m_elf->GetKernelVirtualAddress(kernel_name);
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_HW_BREAKPOINT_SET, &bp_address), 0);

    // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Wait when HW breakpoint is triggered.
    DebugPoll(debug_fd);

    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_HW_BREAKPOINT);

    int step = 0;
    const int MAX_STEPS = 20;

    while (step < MAX_STEPS) {
        // Execute 1 step
        uint32_t do_steps = 1;
        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_STEP, &do_steps), 0);

        // Wait when breakpoint is triggered.
        DebugPoll(debug_fd);

        // Check STOP reason
        ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);

        step += do_steps;

        // Check if STOP executed
        if (state.reason == ELCORE50_STOP_REASON_APP_EXCEPTION) break;
    }

    ASSERT_EQ(step, 6);

    // Detach
    DetachJob(debug_fd);
    WaitJob();
}

TEST_P(JobTests, CheckDbgStepAfterStop) {
    const char kernel_name[255] = "_kernel_empty";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);
    DebugPoll(debug_fd);

    // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Wait for event
    DebugPoll(debug_fd);

    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_APP_EXCEPTION);

    struct elcore50_dbg_mem pc_mem;
    uint32_t pc_old = 0, pc_new = 1;

    pc_mem.vaddr = 0x300;
    pc_mem.size = sizeof(uint32_t);
    pc_mem.data = &pc_old;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_REGISTER_READ, &pc_mem), 0);

    // Execute 1 step
    uint32_t do_steps = 1;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_STEP, &do_steps), 0);

    // Wait for event
    DebugPoll(debug_fd);

    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_APP_EXCEPTION);

    pc_mem.vaddr = 0x300;
    pc_mem.size = sizeof(uint32_t);
    pc_mem.data = &pc_new;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_REGISTER_READ, &pc_mem), 0);

    ASSERT_EQ(pc_new, pc_old);

    // Detach
    DetachJob(debug_fd);
    WaitJob();
}

TEST_P(JobTests, CheckDbgContinueAfterStop) {
    const char kernel_name[255] = "_kernel_empty";
    // Enable debug_enable to attach job
    CreateJobInstance(kernel_name, 1);

    long job_instance_id = WaitJobInstanceCreation(kernel_name, 120);
    ASSERT_NE(job_instance_id, 0);

    int debug_fd = AttachJob(job_instance_id);
    ASSERT_NE(debug_fd, 0);
    DebugPoll(debug_fd);

    // Run DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Wait for event
    DebugPoll(debug_fd);

    struct elcore50_dbg_stop_reason state;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_APP_EXCEPTION);

    struct elcore50_dbg_mem pc_mem;
    uint32_t pc_old = 0, pc_new = 1;

    pc_mem.vaddr = 0x300;
    pc_mem.size = sizeof(uint32_t);
    pc_mem.data = &pc_old;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_REGISTER_READ, &pc_mem), 0);

    // Try to RUN DSP
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_JOB_INSTANCE_CONTINUE, NULL), 0);

    // Wait for event
    DebugPoll(debug_fd);

    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_GET_STOP_REASON, &state), 0);
    ASSERT_EQ(state.reason, ELCORE50_STOP_REASON_APP_EXCEPTION);

    pc_mem.vaddr = 0x300;
    pc_mem.size = sizeof(uint32_t);
    pc_mem.data = &pc_new;
    ASSERT_EQ(ioctl(debug_fd, ELCORE50_IOC_DBG_REGISTER_READ, &pc_mem), 0);

    ASSERT_EQ(pc_new, pc_old);

    // Detach
    DetachJob(debug_fd);
    WaitJob();
}

INSTANTIATE_TEST_CASE_P(GroupTests, JobTests,
                testing::Range(0, static_cast<int>(g_elcore_counter.m_elcore_num)));
