// Copyright 2018 RnD Center "ELVEES", JSC

#include <fstream>
#include <stdexcept>
#include <vector>

#include <fcntl.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>

#include <gtest/gtest.h>
#include <elcorecl/elcorecl.h>

#include "dmabuf-importer.hpp"

#define SZ_1G 0x40000000UL
#define SZ_1K 0x400
#define SZ_1M 0x100000

TEST(Platform, GetNumOfPlatforms) {
    ecl_uint num_of_platforms = 0;
    ASSERT_EQ(eclGetPlatformIDs(0, nullptr, &num_of_platforms), ECL_SUCCESS);
    ASSERT_EQ(num_of_platforms, 1);
}

TEST(Platform, GetPlatformIDs) {
    ecl_platform_id platform_ids[1];
    ASSERT_EQ(eclGetPlatformIDs(1, platform_ids, nullptr), ECL_SUCCESS);
    ASSERT_TRUE(platform_ids[0] != nullptr);
}

TEST(Devices, GetNumOfDevices) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_uint num_devices;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 0, nullptr, &num_devices),
              ECL_SUCCESS);
    ASSERT_GT(num_devices, 0);
}

TEST(Devices, GetDevice) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id device;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 1, &device, nullptr),
              ECL_SUCCESS);
    ASSERT_TRUE(device != nullptr);
}

TEST(Context, Create) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id device;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 1, &device, nullptr),
              ECL_SUCCESS);
    ecl_int result;
    ecl_context context = eclCreateContext(nullptr, 1, &device, nullptr, nullptr, &result);
    ASSERT_TRUE(context != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);
    ASSERT_EQ(eclReleaseContext(context), ECL_SUCCESS);
}

TEST(CommandQueue, Create) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id device;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 1, &device, nullptr),
              ECL_SUCCESS);
    ecl_int result;
    ecl_context context = eclCreateContext(nullptr, 1, &device, nullptr, nullptr, &result);
    ASSERT_TRUE(context != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_command_queue queue =
        eclCreateCommandQueueWithProperties(context, device, nullptr, &result);
    ASSERT_TRUE(queue != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    // it should be possible to create two queues on the same device
    ecl_command_queue queue2 =
        eclCreateCommandQueueWithProperties(context, device, nullptr, &result);
    ASSERT_TRUE(queue2 != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_EQ(eclReleaseCommandQueue(queue), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseCommandQueue(queue2), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseContext(context), ECL_SUCCESS);
}

TEST(Buffer, Create) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id device;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 1, &device, nullptr),
              ECL_SUCCESS);
    ecl_int result;
    ecl_context context = eclCreateContext(nullptr, 1, &device, nullptr, nullptr, &result);
    ASSERT_TRUE(context != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    void *not_aligned_buffer = malloc(140);
    // can't allocate memory without host_ptr
    ecl_mem mem = eclCreateBuffer(context, 0, 140, nullptr, &result);
    ASSERT_TRUE(mem == nullptr);
    ASSERT_EQ(result, ECL_INVALID_HOST_PTR);

    mem = eclCreateBuffer(context, ECL_MEM_USE_HOST_PTR, 140, not_aligned_buffer, &result);
    ASSERT_TRUE(mem == nullptr);
    ASSERT_EQ(result, ECL_INVALID_HOST_PTR);

    free(not_aligned_buffer);

    const int page_size = getpagesize();
    void *aligned_buffer;
    ASSERT_EQ(posix_memalign(&aligned_buffer, page_size, page_size * 2), 0);
    ASSERT_TRUE(aligned_buffer != nullptr);
    mem = eclCreateBuffer(context, ECL_MEM_USE_HOST_PTR, page_size * 2, aligned_buffer, &result);
    ASSERT_TRUE(mem != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    free(aligned_buffer);
    ASSERT_EQ(eclReleaseMemObject(mem), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseContext(context), ECL_SUCCESS);
}

TEST(Buffer, CreateFromDmaBuf) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id device;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 1, &device, nullptr),
              ECL_SUCCESS);
    ecl_int result;
    ecl_context context = eclCreateContext(nullptr, 1, &device, nullptr, nullptr, &result);
    ASSERT_TRUE(context != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    size_t dbuf_size = 2 * getpagesize();
    DmaBuf dbuf(dbuf_size);

    // Try create ecl_mem with size = 0
    ecl_mem mem = eclCreateBufferFromDmaBuf(context, 0, dbuf.get_fd(), 0, &result);
    ASSERT_TRUE(mem == nullptr);
    ASSERT_EQ(result, ECL_INVALID_BUFFER_SIZE);

    // Try create ecl_mem with fd = 0
    mem = eclCreateBufferFromDmaBuf(context, 0, 0, dbuf_size, &result);
    ASSERT_TRUE(mem == nullptr);
    ASSERT_EQ(result, ECL_INVALID_VALUE);

    mem = eclCreateBufferFromDmaBuf(context, 0, dbuf.get_fd(), dbuf_size, &result);
    ASSERT_TRUE(mem != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_EQ(eclReleaseMemObject(mem), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseContext(context), ECL_SUCCESS);
}

TEST(Program, Create) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id device;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 1, &device, nullptr),
              ECL_SUCCESS);
    ecl_int result;
    ecl_context context = eclCreateContext(nullptr, 1, &device, nullptr, nullptr, &result);
    ASSERT_TRUE(context != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    std::vector<char> elf_buffer;
    size_t elf_size;
    {
        std::ifstream file(TEST_ELF_FILE, std::ios::binary | std::ios::ate);
        elf_size = file.tellg();
        file.seekg(0, std::ios::beg);
        elf_buffer.resize(elf_size);
        ASSERT_TRUE(file.read(elf_buffer.data(), elf_size));
    }

    const unsigned char *elfs[1] = {reinterpret_cast<unsigned char *>(elf_buffer.data())};
    ecl_program program =
        eclCreateProgramWithBinary(context, 1, &device, &elf_size, elfs, nullptr, &result);
    ASSERT_TRUE(program != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_EQ(eclReleaseProgram(program), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseContext(context), ECL_SUCCESS);
}

TEST(Program, CreateProgramForTwoDevices) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id devices[2];
    ecl_uint available_devices;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 2, devices, &available_devices),
              ECL_SUCCESS);
    if (available_devices < 2)
        GTEST_SKIP() << "Skip test as the current platform has the only DSP device";
    ecl_int result;
    ecl_context context = eclCreateContext(nullptr, 2, devices, nullptr, nullptr, &result);
    ASSERT_TRUE(context != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    std::vector<char> elf_buffer;
    size_t elf_sizes[2];
    {
        std::ifstream file(TEST_ELF_FILE, std::ios::binary | std::ios::ate);
        elf_sizes[0] = file.tellg();
        file.seekg(0, std::ios::beg);
        elf_buffer.resize(elf_sizes[0]);
        ASSERT_TRUE(file.read(elf_buffer.data(), elf_sizes[0]));
    }

    const unsigned char *elfs[2];
    elfs[0] = elfs[1] = {reinterpret_cast<unsigned char *>(elf_buffer.data())};
    elf_sizes[1] = elf_sizes[0];
    ecl_program program =
        eclCreateProgramWithBinary(context, 2, devices, elf_sizes, elfs, nullptr, &result);
    ASSERT_TRUE(program != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_EQ(eclReleaseProgram(program), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseContext(context), ECL_SUCCESS);
}

TEST(Program, CreateWithSharedSection) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id device;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 1, &device, nullptr),
              ECL_SUCCESS);
    ecl_int result;
    ecl_context context = eclCreateContext(nullptr, 1, &device, nullptr, nullptr, &result);
    ASSERT_TRUE(context != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    std::vector<char> elf_buffer;
    size_t elf_size;
    {
        std::ifstream file(TEST_ELF_FILE, std::ios::binary | std::ios::ate);
        elf_size = file.tellg();
        file.seekg(0, std::ios::beg);
        elf_buffer.resize(elf_size);
        ASSERT_TRUE(file.read(elf_buffer.data(), elf_size));
    }

    size_t shared_section_size = 512 * SZ_1M;
    void *shared_section;
    ecl_mem shared_section_mem;
    ASSERT_EQ(posix_memalign(&shared_section, getpagesize(), shared_section_size), 0);
    ASSERT_TRUE(shared_section != nullptr);
    shared_section_mem = eclCreateBuffer(context, ECL_MEM_USE_HOST_PTR, shared_section_size,
                                         shared_section, &result);
    ASSERT_TRUE(shared_section_mem != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_shared_section section;
    section.mem = shared_section_mem;
    section.virtual_addr = 0x30000000;

    const unsigned char *elfs[1] = {reinterpret_cast<unsigned char *>(elf_buffer.data())};
    ecl_program program =
        eclCreateProgramWithSharedSections(context, 1, &device, &elf_size, elfs, nullptr,
                                           1, &section, &result);
    ASSERT_TRUE(program != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_EQ(eclReleaseMemObject(shared_section_mem), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseProgram(program), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseContext(context), ECL_SUCCESS);
}

TEST(Kernel, Create) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id device;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 1, &device, nullptr),
              ECL_SUCCESS);
    ecl_int result;
    ecl_context context = eclCreateContext(nullptr, 1, &device, nullptr, nullptr, &result);
    ASSERT_TRUE(context != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    std::vector<char> elf_buffer;
    size_t elf_size;
    {
        std::ifstream file(TEST_ELF_FILE, std::ios::binary | std::ios::ate);
        elf_size = file.tellg();
        file.seekg(0, std::ios::beg);
        elf_buffer.resize(elf_size);
        ASSERT_TRUE(file.read(elf_buffer.data(), elf_size));
    }

    const unsigned char *elfs[1] = {reinterpret_cast<unsigned char *>(elf_buffer.data())};
    ecl_program program =
        eclCreateProgramWithBinary(context, 1, &device, &elf_size, elfs, nullptr, &result);
    ASSERT_TRUE(program != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_kernel kernel = eclCreateKernel(program, "_test_pram_add", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseProgram(program), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseContext(context), ECL_SUCCESS);
}

TEST(Kernel, EnqueueWithSharedSection) {
    ecl_platform_id platform_id;
    ASSERT_EQ(eclGetPlatformIDs(1, &platform_id, nullptr), ECL_SUCCESS);
    ecl_device_id device;
    ASSERT_EQ(eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 1, &device, nullptr),
              ECL_SUCCESS);
    ecl_int result;
    ecl_context context = eclCreateContext(nullptr, 1, &device, nullptr, nullptr, &result);
    ASSERT_TRUE(context != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    std::vector<char> elf_buffer;
    size_t elf_size;
    {
        std::ifstream file(TEST_ELF_FILE, std::ios::binary | std::ios::ate);
        elf_size = file.tellg();
        file.seekg(0, std::ios::beg);
        elf_buffer.resize(elf_size);
        ASSERT_TRUE(file.read(elf_buffer.data(), elf_size));
    }

    const size_t shared_section_size = 512 * SZ_1M;
    void *shared_section;
    ecl_mem shared_section_mem;
    ASSERT_EQ(posix_memalign(&shared_section, getpagesize(), shared_section_size), 0);
    ASSERT_TRUE(shared_section != nullptr);
    shared_section_mem = eclCreateBuffer(context, ECL_MEM_USE_HOST_PTR, shared_section_size,
                                         shared_section, &result);
    ASSERT_TRUE(shared_section_mem != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_shared_section section;
    section.mem = shared_section_mem;
    section.virtual_addr = 0x30000000;

    const unsigned char *elfs[1] = {reinterpret_cast<unsigned char *>(elf_buffer.data())};
    ecl_program program =
        eclCreateProgramWithSharedSections(context, 1, &device, &elf_size, elfs, nullptr,
                                           1, &section, &result);
    ASSERT_TRUE(program != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_kernel kernel = eclCreateKernel(program, "_fill_array", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    uint8_t etalon_value = 0xBA;

    ecl_command_queue queue = eclCreateCommandQueueWithProperties(context, device,
                                                                  nullptr, &result);
    ASSERT_TRUE(queue != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_EQ(eclSetKernelArg(kernel, 0, sizeof(ecl_long), &section.virtual_addr), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArg(kernel, 1, sizeof(size_t), &shared_section_size), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArg(kernel, 2, sizeof(int32_t), &etalon_value), ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    unsigned char *etalon_array = (unsigned char *) malloc(shared_section_size);
    ASSERT_TRUE(etalon_array != nullptr);
    memset(etalon_array, etalon_value, shared_section_size);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(queue, shared_section_mem, ECL_TRUE, ECL_MAP_READ, 0, shared_section_size,
                        0, NULL, NULL, &result);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_FALSE(memcmp(shared_section, etalon_array, shared_section_size));
    free(etalon_array);
    free(shared_section);

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(shared_section_mem), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseCommandQueue(queue), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseProgram(program), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseContext(context), ECL_SUCCESS);
}

class DeviceCounter {
 public:
    DeviceCounter() {
        ecl_platform_id platform_id;
        if (eclGetPlatformIDs(1, &platform_id, nullptr) != ECL_SUCCESS) {
            throw std::runtime_error("Can't get platform id");
        }
        if (eclGetDeviceIDs(platform_id, ECL_DEVICE_TYPE_CUSTOM, 0, nullptr, &m_devices_num) !=
            ECL_SUCCESS) {
            throw std::runtime_error("Can't get device ids");
        }
    }
    ecl_uint m_devices_num;
};

DeviceCounter g_dev_counter;

void ECL_CALLBACK MemoryDestructor(ecl_mem, void *user_data) { free(user_data); }

class EnqueuedCommands : public ::testing::TestWithParam<int> {
 public:
    EnqueuedCommands() : m_context(nullptr), m_program(nullptr), m_queue(nullptr) {}

 protected:
    void SetUp() override {
        m_core_num = GetParam();
        ASSERT_EQ(eclGetPlatformIDs(1, &m_platform_id, nullptr), ECL_SUCCESS);
        std::vector<ecl_device_id> devices(g_dev_counter.m_devices_num);
        ASSERT_EQ(eclGetDeviceIDs(m_platform_id, ECL_DEVICE_TYPE_CUSTOM,
                                  g_dev_counter.m_devices_num, &devices[0], nullptr),
                  ECL_SUCCESS);
        m_device = devices[m_core_num];

        ecl_int result;
        m_context = eclCreateContext(nullptr, 1, &m_device, nullptr, nullptr, &result);
        ASSERT_TRUE(m_context != nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);

        std::vector<char> elf_buffer;
        size_t elf_size;
        {
            std::ifstream file(TEST_ELF_FILE, std::ios::binary | std::ios::ate);
            elf_size = file.tellg();
            file.seekg(0, std::ios::beg);
            elf_buffer.resize(elf_size);
            ASSERT_TRUE(file.read(elf_buffer.data(), elf_size));
        }

        const unsigned char *elfs[1] = {reinterpret_cast<unsigned char *>(elf_buffer.data())};
        m_program =
            eclCreateProgramWithBinary(m_context, 1, &m_device, &elf_size, elfs, nullptr, &result);
        ASSERT_TRUE(m_program != nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);

        m_queue = eclCreateCommandQueueWithProperties(m_context, m_device, nullptr, &result);
        ASSERT_TRUE(m_queue != nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);
    }

    void TearDown() override {
        if (m_queue) {
            ASSERT_EQ(eclReleaseCommandQueue(m_queue), ECL_SUCCESS);
        }
        if (m_program) {
            ASSERT_EQ(eclReleaseProgram(m_program), ECL_SUCCESS);
        }
        if (m_context) {
            ASSERT_EQ(eclReleaseContext(m_context), ECL_SUCCESS);
        }
    }

    void CreateBuffer(size_t size, ecl_mem &mem, void **p,
                      std::function<void(void *ptr, size_t size)> fill = nullptr) {
        ecl_int result;
        const int page_size = getpagesize();
        size = ((size + page_size - 1) / page_size) * page_size;
        ASSERT_EQ(posix_memalign(p, page_size, size), 0);
        ASSERT_TRUE(*p != nullptr);
        if (fill)
            fill(*p, size);
        mem = eclCreateBuffer(m_context, ECL_MEM_USE_HOST_PTR, size, *p, &result);
        ASSERT_TRUE(mem != nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);

        ASSERT_EQ(eclSetMemObjectDestructorCallback(mem, MemoryDestructor, *p), ECL_SUCCESS);
    }

    void CreateDMABuffer(size_t size, ecl_mem &mem, void **p,
                         std::function<void(void *ptr, size_t size)> fill = nullptr) {
        ecl_int result;
        const int page_size = getpagesize();
        size = ((size + page_size - 1) / page_size) * page_size;
        ASSERT_EQ(posix_memalign(p, page_size, size), 0);
        ASSERT_TRUE(*p != nullptr);
        if (fill)
            fill(*p, size);
        mem = eclCreateDMABuffer(m_context, ECL_MEM_USE_HOST_PTR, size, *p, &result);
        ASSERT_TRUE(mem != nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);

        ASSERT_EQ(eclSetMemObjectDestructorCallback(mem, MemoryDestructor, *p), ECL_SUCCESS);
    }

    // Avoid use this function twice in one test with the same 'expected_result' parameter because
    // it will difficult to find which call caused assert.
    void RunKernelAndCheckWaitResult(const char *const kernel_name, ecl_int expected_result) {
        ecl_int result;

        ecl_kernel kernel = eclCreateKernel(m_program, kernel_name, &result);
        ASSERT_TRUE(kernel != nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);

        const size_t global_work_size[1] = {1};
        ecl_event kernel_event;
        ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                          nullptr, &kernel_event),
                  ECL_SUCCESS);
        ASSERT_EQ(eclWaitForEvents(1, &kernel_event), expected_result);

        ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
        ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
    }

    void RunCheckSyscallGettimeofday(const char *const kernel_name) {
        ecl_int result;
        struct timeval tv_start, tv_end, *tv_sys;

        ecl_kernel kernel = eclCreateKernel(m_program, kernel_name, &result);
        ASSERT_TRUE(kernel != nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);

        ecl_mem res;
        CreateBuffer(sizeof(timeval), res, reinterpret_cast<void **>(&tv_sys));
        ASSERT_EQ(eclSetKernelArgELcoreMem(kernel, 0, res), ECL_SUCCESS);

        const size_t global_work_size[1] = {1};
        ecl_event kernel_event;
        gettimeofday(&tv_start, NULL);
        ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                          nullptr, &kernel_event),
                  ECL_SUCCESS);
        ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);
        gettimeofday(&tv_end, NULL);

        eclEnqueueMapBuffer(m_queue, res, ECL_TRUE, ECL_MAP_READ, 0, sizeof(timeval), 0, NULL,
                            NULL, &result);
        ASSERT_EQ(result, ECL_SUCCESS);

        // Check that tv_start <= tv_sys <= tv_end
        ASSERT_EQ((timercmp(&tv_start, tv_sys, >) | timercmp(tv_sys, &tv_end, >)), false);

        ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
        ASSERT_EQ(eclReleaseMemObject(res), ECL_SUCCESS);
        ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
    }

    ecl_platform_id m_platform_id;
    ecl_uint m_core_num;
    ecl_device_id m_device;
    ecl_context m_context;
    ecl_program m_program;
    ecl_command_queue m_queue;
};

TEST_P(EnqueuedCommands, QueueAndWaitOneKernel) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_add", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_mem res;
    int *z;
    int x = 30, y = 40;
    CreateBuffer(sizeof(int), res, reinterpret_cast<void **>(&z),
                 [](void *ptr, int size) {
                     int *z = (int *) ptr;
                     *z = 80;
                 });

    ASSERT_EQ(eclSetKernelArg(kernel, 0, sizeof(int), reinterpret_cast<void *>(&x)), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArg(kernel, 1, sizeof(int), reinterpret_cast<void *>(&y)), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgELcoreMem(kernel, 2, res), ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, res, ECL_TRUE, ECL_MAP_READ, 0, sizeof(int), 0, NULL, NULL,
                        &result);
    ASSERT_EQ(result, ECL_SUCCESS);

    EXPECT_EQ(x + y, *z);

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, QueueOneKernelWithWaitList) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_add", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_mem res;
    int x = 30, y = 40;
    int *z;
    CreateBuffer(sizeof(int), res, reinterpret_cast<void **>(&z),
                 [](void *ptr, int size) {
                     int *z = (int *) ptr;
                     *z = 80;
                 });
    ASSERT_EQ(eclSetKernelArg(kernel, 0, sizeof(int), reinterpret_cast<void *>(&x)), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArg(kernel, 1, sizeof(int), reinterpret_cast<void *>(&y)), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgELcoreMem(kernel, 2, res), ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ecl_event user_event = eclCreateUserEvent(m_context, &result);
    ASSERT_TRUE(user_event != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 1,
                                      &user_event, &kernel_event),
              ECL_SUCCESS);
    eclSetUserEventStatus(user_event, ECL_COMPLETE);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, res, ECL_TRUE, ECL_MAP_READ, 0, sizeof(int), 0, NULL, NULL,
                        &result);
    ASSERT_EQ(result, ECL_SUCCESS);

    EXPECT_EQ(x + y, *z);

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseEvent(user_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckConstructor) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_get_z", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_mem res;
    int *z;
    CreateBuffer(sizeof(int), res, reinterpret_cast<void **>(&z),
                 [](void *ptr, int size) {
                     int *z = (int *) ptr;
                     *z = 5;
                 });
    ASSERT_EQ(eclSetKernelArgELcoreMem(kernel, 0, res), ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, res, ECL_TRUE, ECL_MAP_READ, 0, sizeof(int), 0, NULL, NULL,
                        &result);
    ASSERT_EQ(result, ECL_SUCCESS);

    EXPECT_EQ(30, *z);

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

typedef struct __attribute__((packed)) {
    int a[3];
} stack_struct;

TEST_P(EnqueuedCommands, CheckStack) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_stack", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    stack_struct src;
    src.a[0] = 0;
    src.a[1] = 1;
    src.a[2] = 2;
    ASSERT_EQ(eclSetKernelArg(kernel, 0, sizeof(stack_struct), &src), ECL_SUCCESS);

    ecl_mem res;
    stack_struct *dst;
    CreateBuffer(sizeof(stack_struct), res, reinterpret_cast<void **>(&dst),
                 [](void *ptr, int size) {
                     stack_struct *dst = (stack_struct *) ptr;
                     dst->a[0] = dst->a[1] = dst->a[2] = -1;
                 });
    ASSERT_EQ(eclSetKernelArgELcoreMem(kernel, 1, res), ECL_SUCCESS);
    ASSERT_EQ(result, ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, res, ECL_TRUE, ECL_MAP_READ, 0, sizeof(stack_struct), 0, NULL,
                        NULL, &result);
    ASSERT_EQ(result, ECL_SUCCESS);

    EXPECT_EQ(dst->a[0], src.a[0]);
    EXPECT_EQ(dst->a[1], src.a[1]);
    EXPECT_EQ(dst->a[2], src.a[2]);

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, DebugBreak) {
    RunKernelAndCheckWaitResult("_test_debug_break", ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckTLBReadExceptions) {
#ifdef BUILD_FOR_VP
    if (m_core_num % 4)
        GTEST_SKIP() << "Skip test as rf#12331 is not fixed";
#endif
    RunKernelAndCheckWaitResult("_test_tlb_rex", ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckTLBWriteExceptions) {
#ifdef BUILD_FOR_VP
    if (m_core_num % 4)
        GTEST_SKIP() << "Skip test as rf#12331 is not fixed";
#endif
    RunKernelAndCheckWaitResult("_test_tlb_wex", ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, ZeroIntDivision) {
#ifdef BUILD_FOR_VP
    GTEST_SKIP() << "Skip test as rf#12407 is not fixed";
#endif
    RunKernelAndCheckWaitResult("_test_zero_int_division",
                                ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, ZeroFloatDivision) {
    RunKernelAndCheckWaitResult("_test_zero_float_division", ECL_SUCCESS);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, StopWithTrap) {
    RunKernelAndCheckWaitResult("_test_stop_with_trap",
                                ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, SyscallGettimeofday) {
    RunCheckSyscallGettimeofday("_test_syscall_gettimeofday");
}

TEST_P(EnqueuedCommands, SyscallGettimeofdayDynamic) {
    RunCheckSyscallGettimeofday("_test_syscall_gettimeofday_dynamic");
}

TEST_P(EnqueuedCommands, StopWithUndefinedSyscall) {
    RunKernelAndCheckWaitResult("_test_stop_with_undefined_syscall",
                                ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, UndefinedInstruction) {
    RunKernelAndCheckWaitResult("_test_undefined_instruction",
                                ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, BadAddress) {
#ifdef BUILD_FOR_VP
    GTEST_SKIP() << "Skip test as rf#12408 is not fixed";
#endif
    RunKernelAndCheckWaitResult("_test_bad_address", ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, DoStackException) {
    RunKernelAndCheckWaitResult("_test_do_stack_exception",
                                ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    RunKernelAndCheckWaitResult("_test_should_success_stop", ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, IncrementArrayInXYRAMViaDMA) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_increment_array_in_xyram_via_dma",
                                        &result);
    ASSERT_NE(kernel, nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    size_t array_size = 384 * SZ_1K;

    ecl_mem res;
    unsigned char *array_to_increment, incremented_array[array_size];
    CreateDMABuffer(array_size, res, reinterpret_cast<void **>(&array_to_increment),
                    [](void *ptr, int size) { memset(ptr, 0x71, size); });
    ASSERT_EQ(eclSetKernelArgDMAMem(kernel, 0, res), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgELcoreLocalMem(kernel, 1, array_size), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArg(kernel, 2, sizeof(size_t), reinterpret_cast<void *>(&array_size)),
              ECL_SUCCESS);

    memset(incremented_array, 0x72, array_size);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, res, ECL_TRUE, ECL_MAP_READ, 0, array_size, 0, NULL, NULL,
                        &result);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_FALSE(memcmp(array_to_increment, incremented_array, array_size));

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, SendArrayInXYRAMViaDMAChains) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_send_array_in_xyram_via_dma_chains",
                                        &result);
    ASSERT_NE(kernel, nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    size_t array_size = 512 * SZ_1K;

    ecl_mem res, resout;
    unsigned char *source_array, *dest_array;
    CreateDMABuffer(array_size, res, reinterpret_cast<void **>(&source_array),
                    [](void *ptr, int size) {
                        int fd = open("/dev/urandom", O_RDONLY);
                        ASSERT_NE(fd, -1);
                        ASSERT_EQ(read(fd, ptr, size), size);
                        close(fd);
                    });
    CreateDMABuffer(array_size, resout, reinterpret_cast<void **>(&dest_array));
    ASSERT_EQ(eclSetKernelArgDMAMem(kernel, 0, res), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgELcoreLocalMem(kernel, 1, array_size), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgDMAMem(kernel, 2, resout), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArg(kernel, 3, sizeof(size_t), reinterpret_cast<void *>(&array_size)),
              ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, resout, ECL_TRUE, ECL_MAP_READ, 0, array_size, 0, NULL, NULL,
                        &result);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_FALSE(memcmp(source_array, dest_array, array_size));

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(resout), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckLocalArgsArrange) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_local_args_arrange", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    const int largs = 5;

    ecl_mem res;
    uint32_t *local_args_addr;
    CreateBuffer(largs * sizeof(uint32_t), res, reinterpret_cast<void **>(&local_args_addr));

    size_t sizes[largs] = {16 * SZ_1K, 3 * SZ_1K, 95 * SZ_1K, 180 * SZ_1K,
                           48 * SZ_1K};
    for (int i = 0; i < largs; ++i)
        ASSERT_EQ(eclSetKernelArgELcoreLocalMem(kernel, i, sizes[i]), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgELcoreMem(kernel, largs, res), ECL_SUCCESS);


    uint32_t expected_local_args_addr[largs] = {0x40000, 0x40000 + 16 * SZ_1K,
                                                0x40000 + 19 * SZ_1K, 0,
                                                0x40000 + 114 * SZ_1K};

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, res, ECL_TRUE, ECL_MAP_READ, 0, largs * sizeof(uint32_t), 0, NULL,
                        NULL, &result);
    ASSERT_EQ(result, ECL_SUCCESS);
    ASSERT_FALSE(memcmp(expected_local_args_addr, local_args_addr, largs * sizeof(uint32_t)));

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckNonCrossingElfSectionsAndNonCachedPools) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_should_success_stop", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_mem res;
    int *z;
    // For size = 3.64 G elf section crosses non-cached memory area
    CreateBuffer(3 * SZ_1G + 640 * SZ_1M, res, reinterpret_cast<void **>(&z));
    ASSERT_EQ(eclSetKernelArgELcoreMemNonCached(kernel, 0, res), ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);

    ASSERT_EQ(eclReleaseMemObject(res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckAlienDmaBuffer) {
    if (m_core_num == 0)
        GTEST_SKIP() << "Skip test as dmabuf is created also for device 0";

    ecl_int result;
    ecl_kernel kernel = eclCreateKernel(m_program, "_fill_array", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    size_t dbuf_size = 15000;
    DmaBuf dbuf(dbuf_size);
    int32_t expected_value = 137;
    void *expected_buffer = malloc(dbuf_size);
    memset(expected_buffer, expected_value, dbuf_size);

    ecl_mem mem = eclCreateBufferFromDmaBuf(m_context, 0, dbuf.get_fd(), dbuf_size, &result);
    ASSERT_TRUE(mem != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);
    eclSetKernelArgELcoreMem(kernel, 0, mem);
    eclSetKernelArg(kernel, 1, sizeof(size_t), &dbuf_size);
    eclSetKernelArg(kernel, 2, sizeof(int32_t), &expected_value);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, mem, ECL_TRUE, ECL_MAP_READ | ECL_MAP_WRITE, 0, dbuf_size, 0,
                        NULL, NULL, &result);
    ASSERT_EQ(result, ECL_SUCCESS);
    // Compare results
    ASSERT_FALSE(memcmp(expected_buffer, dbuf.m_data, dbuf_size));

    ASSERT_EQ(eclReleaseMemObject(mem), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckMapUnmapDmaBuffer) {
    ecl_int result;
    ecl_kernel kernel = eclCreateKernel(m_program, "_increment_array", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    size_t dbuf_size = 15000;
    int fill_value = 56;
    DmaBuf dbuf(dbuf_size);
    void *expected_buffer = malloc(dbuf_size);
    void *mapped_ptr;

    ecl_mem mem = eclCreateBufferFromDmaBuf(m_context, 0, dbuf.get_fd(), dbuf_size, &result);
    ASSERT_TRUE(mem != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);
    eclSetKernelArgELcoreMem(kernel, 0, mem);
    eclSetKernelArg(kernel, 1, sizeof(size_t), &dbuf_size);

    // Check blocking map
    mapped_ptr = eclEnqueueMapBuffer(m_queue, mem, ECL_TRUE, ECL_MAP_WRITE, 0, dbuf_size, 0, NULL,
                                     NULL, &result);
    ASSERT_TRUE(mapped_ptr != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    memset(dbuf.m_data, fill_value, dbuf_size);
    memset(expected_buffer, fill_value + 1, dbuf_size);

    const size_t global_work_size[1] = {1};
    ecl_event event;

    // Check unmap
    ASSERT_EQ(eclEnqueueUnmapMemObject(m_queue, mem, mapped_ptr, 0, NULL, &event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &event), ECL_SUCCESS);

    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &event), ECL_SUCCESS);

    // Check unblocking map
    mapped_ptr = eclEnqueueMapBuffer(m_queue, mem, ECL_FALSE, ECL_MAP_WRITE, 0, dbuf_size, 0, NULL,
                                     &event, &result);
    ASSERT_EQ(result, ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &event), ECL_SUCCESS);
    ASSERT_TRUE(mapped_ptr != nullptr);

    // Compare results
    ASSERT_FALSE(memcmp(expected_buffer, dbuf.m_data, dbuf_size));

    ASSERT_EQ(eclReleaseMemObject(mem), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckMapUnmapBuffer) {
    ecl_int result;
    ecl_kernel kernel = eclCreateKernel(m_program, "_increment_array", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    size_t buf_size = 15000;
    int fill_value = 56;
    void *expected_buffer = malloc(buf_size);
    void *mapped_ptr;

    void *buffer; ecl_mem mem;
    CreateBuffer(buf_size, mem, &buffer);
    ASSERT_TRUE(mem != nullptr);
    eclSetKernelArgELcoreMem(kernel, 0, mem);
    eclSetKernelArg(kernel, 1, sizeof(size_t), &buf_size);

    // Check blocking map
    mapped_ptr = eclEnqueueMapBuffer(m_queue, mem, ECL_TRUE, ECL_MAP_WRITE, 0, buf_size, 0, NULL,
                                     NULL, &result);
    ASSERT_TRUE(mapped_ptr != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    memset(buffer, fill_value, buf_size);
    memset(expected_buffer, fill_value + 1, buf_size);

    const size_t global_work_size[1] = {1};
    ecl_event event;

    // Check unmap
    ASSERT_EQ(eclEnqueueUnmapMemObject(m_queue, mem, mapped_ptr, 0, NULL, &event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &event), ECL_SUCCESS);

    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &event), ECL_SUCCESS);

    // Check unblocking map
    mapped_ptr = eclEnqueueMapBuffer(m_queue, mem, ECL_FALSE, ECL_MAP_WRITE, 0, buf_size, 0, NULL,
                                     &event, &result);
    ASSERT_EQ(result, ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &event), ECL_SUCCESS);
    ASSERT_TRUE(mapped_ptr != nullptr);

    // Compare results
    ASSERT_FALSE(memcmp(expected_buffer, buffer, buf_size));

    ASSERT_EQ(eclReleaseMemObject(mem), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckEnqueueMultipleCommands) {
    ecl_int result;
    ecl_kernel kernel = eclCreateKernel(m_program, "_increment_array", &result);
    ASSERT_TRUE(kernel != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    size_t buf_size = 15000;
    int fill_value = 56;
    void *expected_buffer = malloc(buf_size);
    void *mapped_ptr;

    void *buffer; ecl_mem mem;
    CreateBuffer(buf_size, mem, &buffer);
    ASSERT_TRUE(mem != nullptr);
    eclSetKernelArgELcoreMem(kernel, 0, mem);
    eclSetKernelArg(kernel, 1, sizeof(size_t), &buf_size);

    // Check blocking map
    mapped_ptr = eclEnqueueMapBuffer(m_queue, mem, ECL_TRUE, ECL_MAP_WRITE, 0, buf_size, 0, NULL,
                                     NULL, &result);
    ASSERT_TRUE(mapped_ptr != nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    memset(buffer, fill_value, buf_size);
    memset(expected_buffer, fill_value + 1, buf_size);

    const size_t global_work_size[1] = {1};
    ecl_event event;

    // The first command
    eclEnqueueUnmapMemObject(m_queue, mem, mapped_ptr, 0, NULL, &event);
    // The second command
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &event),
              ECL_SUCCESS);
    // The third command
    mapped_ptr = eclEnqueueMapBuffer(m_queue, mem, ECL_FALSE, ECL_MAP_WRITE, 0, buf_size, 0, NULL,
                                     &event, &result);
    ASSERT_EQ(result, ECL_SUCCESS);
    // Waiting for competion of three commands
    ASSERT_EQ(eclWaitForEvents(1, &event), ECL_SUCCESS);
    ASSERT_TRUE(mapped_ptr != nullptr);

    // Compare results
    ASSERT_FALSE(memcmp(expected_buffer, buffer, buf_size));

    ASSERT_EQ(eclReleaseMemObject(mem), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckFileSystemCalls) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_file_systemcalls", &result);
    ASSERT_NE(kernel, nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    size_t string_size = 15;

    ecl_mem source_res, dest_res;
    char *source_string, *dest_string;
    CreateBuffer(string_size, source_res, reinterpret_cast<void **>(&source_string),
                    [](void *ptr, int size) { sprintf((char *)ptr, "Testing message"); });
    CreateBuffer(string_size, dest_res, reinterpret_cast<void **>(&dest_string));

    ASSERT_EQ(eclSetKernelArg(kernel, 0, sizeof(ecl_uint), &m_core_num), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgELcoreMem(kernel, 1, source_res), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgELcoreMem(kernel, 2, dest_res), ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, dest_res, ECL_TRUE, ECL_MAP_READ, 0, string_size, 0, NULL, NULL,
                        &result);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_FALSE(memcmp(dest_string, source_string, string_size));

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(dest_res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(source_res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedCommands, CheckCorrectAccessVirtualAddresses) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_check_virtual_addresses",
                                        &result);
    ASSERT_NE(kernel, nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    int stage = 0; ecl_mem res;
    char *buf;
    CreateBuffer(2 * 4096, res, reinterpret_cast<void **>(&buf));
    ASSERT_EQ(eclSetKernelArg(kernel, 0, sizeof(int), &stage), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgELcoreMem(kernel, 1, res), ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(res), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);

    for (stage = 1; stage <= 2; stage++) {
        ecl_kernel kernel = eclCreateKernel(m_program, "_test_check_virtual_addresses",
                                            &result);
        ASSERT_NE(kernel, nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);
        ASSERT_EQ(eclSetKernelArg(kernel, 0, sizeof(int), &stage), ECL_SUCCESS);
        ecl_event kernel_event;
        ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                          nullptr, &kernel_event),
                  ECL_SUCCESS);
        ASSERT_EQ(eclWaitForEvents(1, &kernel_event),
                  ECL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST);
        ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
        ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
    }
}

TEST_P(EnqueuedCommands, CheckNonCachedArguments) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_nc_arguments",
                                        &result);
    ASSERT_NE(kernel, nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_mem res_in, res_out;
    size_t size = 32 * SZ_1K;
    char *buf_in, *buf_out;
    CreateBuffer(size, res_in, reinterpret_cast<void **>(&buf_in),
                    [](void *ptr, int size) {
                        int fd = open("/dev/urandom", O_RDONLY);
                        ASSERT_NE(fd, -1);
                        ASSERT_EQ(read(fd, ptr, size), size);
                        close(fd);
                    });
    CreateBuffer(size, res_out, reinterpret_cast<void **>(&buf_out));
    ASSERT_EQ(eclSetKernelArgELcoreMemNonCached(kernel, 0, res_in), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArgELcoreMemNonCached(kernel, 1, res_out), ECL_SUCCESS);
    ASSERT_EQ(eclSetKernelArg(kernel, 2, sizeof(size_t), &size), ECL_SUCCESS);

    const size_t global_work_size[1] = {1};
    ecl_event kernel_event;
    ASSERT_EQ(eclEnqueueNDRangeKernel(m_queue, kernel, 1, nullptr, global_work_size, nullptr, 0,
                                      nullptr, &kernel_event),
              ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    eclEnqueueMapBuffer(m_queue, res_in, ECL_TRUE, ECL_MAP_READ, 0, size, 0, NULL, NULL,
                        &result);
    ASSERT_EQ(result, ECL_SUCCESS);
    eclEnqueueMapBuffer(m_queue, res_out, ECL_TRUE, ECL_MAP_READ, 0, size, 0, NULL, NULL,
                        &result);
    ASSERT_EQ(result, ECL_SUCCESS);

    ASSERT_FALSE(memcmp(buf_in, buf_out, size));

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(res_in), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseMemObject(res_out), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

class EnqueuedSpawnCommands : public ::testing::TestWithParam<int> {
 public:
    EnqueuedSpawnCommands() : m_devices(g_dev_counter.m_devices_num),
                              m_context(nullptr), m_program(nullptr), m_queue(nullptr) {}

 protected:
    void SetUp() override {
        m_ncores = GetParam();
        ASSERT_EQ(eclGetPlatformIDs(1, &m_platform_id, nullptr), ECL_SUCCESS);
        ASSERT_EQ(eclGetDeviceIDs(m_platform_id, ECL_DEVICE_TYPE_CUSTOM, m_ncores, &m_devices[0],
                                  nullptr),
                  ECL_SUCCESS);

        ecl_int result;
        m_context = eclCreateContext(nullptr, m_ncores, &m_devices[0], nullptr, nullptr, &result);
        ASSERT_NE(m_context, nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);

        std::vector<char> elf_buffer;
        size_t elf_sizes[m_ncores];
        {
            std::ifstream file(TEST_ELF_FILE, std::ios::binary | std::ios::ate);
            elf_sizes[0] = file.tellg();
            file.seekg(0, std::ios::beg);
            elf_buffer.resize(elf_sizes[0]);
            ASSERT_TRUE(file.read(elf_buffer.data(), elf_sizes[0]));
        }

        const unsigned char *elfs[m_ncores];
        for (ecl_uint i = 0; i < m_ncores; ++i) {
            elf_sizes[i] = elf_sizes[0];
            elfs[i] = reinterpret_cast<unsigned char *>(elf_buffer.data());
        }

        m_program = eclCreateProgramWithBinary(m_context, m_ncores, &m_devices[0], elf_sizes,
                                               elfs, nullptr, &result);
        ASSERT_TRUE(m_program != nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);

        m_queue = eclCreateMultiCommandQueueWithProperties(m_context, m_devices[0], nullptr,
                                                           m_ncores - 1, &m_devices[1], &result);
        ASSERT_TRUE(m_queue != nullptr);
        ASSERT_EQ(result, ECL_SUCCESS);
    }

    void TearDown() override {
        if (m_queue) {
            ASSERT_EQ(eclReleaseMultiCommandQueue(m_queue), ECL_SUCCESS);
        }
        if (m_program) {
            ASSERT_EQ(eclReleaseProgram(m_program), ECL_SUCCESS);
        }
        if (m_context) {
            ASSERT_EQ(eclReleaseContext(m_context), ECL_SUCCESS);
        }
    }

    ecl_int RunKernel(ecl_kernel kernel, ecl_event *event) {
        return eclEnqueueKernelWithSpawn(m_queue, kernel, 0, nullptr, event);
    }

    ecl_platform_id m_platform_id;
    std::vector<ecl_device_id> m_devices;
    ecl_uint m_ncores;
    ecl_context m_context;
    ecl_program m_program;
    ecl_multi_command_queue m_queue;
};

TEST_P(EnqueuedSpawnCommands, TestSpawnConvolution) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_parallel_convolution", &result);
    ASSERT_NE(kernel, nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    ecl_event kernel_event;
    ASSERT_EQ(RunKernel(kernel, &kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclWaitForEvents(1, &kernel_event), ECL_SUCCESS);

    ASSERT_EQ(eclReleaseEvent(kernel_event), ECL_SUCCESS);
    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

TEST_P(EnqueuedSpawnCommands, TestSpawnWithStackArg) {
    ecl_int result;

    ecl_kernel kernel = eclCreateKernel(m_program, "_test_should_success_stop", &result);
    ASSERT_NE(kernel, nullptr);
    ASSERT_EQ(result, ECL_SUCCESS);

    uint64_t stack_argument[2];
    ASSERT_EQ(eclSetKernelArg(kernel, 0, 16, stack_argument), ECL_SUCCESS);

    ASSERT_EQ(RunKernel(kernel, nullptr), ECL_INVALID_KERNEL_ARGS);

    ASSERT_EQ(eclReleaseKernel(kernel), ECL_SUCCESS);
}

INSTANTIATE_TEST_CASE_P(EnqueuedGroup, EnqueuedCommands,
                        testing::Range(0, static_cast<int>(g_dev_counter.m_devices_num)));
#ifdef BUILD_FOR_MCOM03
INSTANTIATE_TEST_CASE_P(EnqueuedGroup, EnqueuedSpawnCommands, testing::Values(2));
#else
INSTANTIATE_TEST_CASE_P(EnqueuedGroup, EnqueuedSpawnCommands, testing::Values(2, 4, 8, 16));
#endif
