// Copyright 2021 RnD Center "ELVEES", JSC

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include <linux/elcore50.h>

static void help(const char *program_name) {
    printf("ELcore control utility\n\n");
    printf("Usage: %s [COMMAND]\n\n", program_name);
    printf("Commands:\n");
    printf("    ps    Report a snapshot of the current ELcore jobs\n");
}

static int get_elcore_count() {
    int elcore_fd;
    struct elcore50_device_info info;

    elcore_fd = open("/dev/elcore0", O_RDWR, 0);
    if (elcore_fd < 0) return 0;

    if (ioctl(elcore_fd, ELCORE50_IOC_GET_CORE_IDX, &info) < 0) return 0;
    close(elcore_fd);
    return info.nclusters * info.cluster_cap;
}

static int elcore_ps() {
    struct elcore50_job_instance_list list;
    char devname[32];
    int ret;
    __u32 job_count;

    printf("%-10s%-10s%-25s%-25s\n", "ID", "PID", "DEVICE", "NAME");
    for (int dev = 0; dev < get_elcore_count(); ++dev) {
        sprintf(devname, "/dev/elcore%i", dev);
        if (access(devname, F_OK ) != 0)
            continue;  // file does not exist
        int fd = open(devname, O_RDWR, 0);
        if (fd < 0) {
            fprintf(stderr, "Could not open device file %s\n", devname);
            return fd;
        }

        ret = ioctl(fd, ELCORE50_IOC_GET_JOB_COUNT, &job_count);
        if (ret < 0) {
            fprintf(stderr, "ELCORE50_IOC_GET_JOB_COUNT failed with %d\n", errno);
            return ret;
        }
        if (!job_count) continue;

        list.info = malloc(job_count * sizeof(struct elcore50_job_instance_info));
        if (!list.info) {
            fprintf(stderr, "Failed to allocate memory\n");
            return ret;
        }
        list.job_instance_count = job_count;
        list.job_instance_ret = 0;

        ret = ioctl(fd, ELCORE50_IOC_GET_JOB_LIST, &list);
        if (ret < 0) {
            free(list.info);
            fprintf(stderr, "ELCORE50_IOC_GET_JOB_LIST failed with %d\n", errno);
            return ret;
        }
        for (int i = 0; i < list.job_instance_ret; ++i) {
            struct elcore50_job_instance_info info = list.info[i];
            printf("%-10lx%-10d%-25s%-25s\n", info.id, info.pid, devname, info.name);
        }
        free(list.info);
        close(fd);
    }
    return 0;
}

int main(int argc, char **argv) {
    int opt;

    while ((opt = getopt(argc, argv, "h")) != -1) {
        switch (opt) {
        case 'h':
            help(argv[0]);
            return 0;
        default:
            fprintf(stderr, "Use %s -h for help.\n", argv[0]);
            return 1;
        }
    }

    if (optind >= argc) {
        fprintf(stderr, "Need command. Use %s -h for help.\n", argv[0]);
        return 1;
    }

    char *command = argv[optind];
    if (!strcmp(command, "ps")) {
        return elcore_ps();
    } else {
        fprintf(stderr, "Wrong command %s. Use %s -h for help.\n", command, argv[0]);
        return 1;
    }
}
